@@ -17,6 +17,7 @@
 
 from __future__ import absolute_import, division, print_function
 
+import math
 from collections import defaultdict
 from enum import IntEnum
 
@@ -33,7 +34,7 @@ from hypothesis.internal.compat import (
     text_type,
     unicode_safe_repr,
 )
-from hypothesis.internal.conjecture.junkdrawer import IntList
+from hypothesis.internal.conjecture.junkdrawer import IntList, uniform
 from hypothesis.internal.conjecture.utils import calc_label_from_name
 from hypothesis.internal.escalation import mark_for_escalation
 
@@ -179,7 +180,7 @@ class Example(object):
 
 class ExampleProperty(object):
     """There are many properties of examples that we calculate by
-    essentially rerunning thet test case multiple times based on the
+    essentially rerunning the test case multiple times based on the
     calls which we record in ExampleRecord.
 
     This class defines a visitor, subclasses of which can be used
@@ -238,7 +239,7 @@ class ExampleProperty(object):
 
     def block(self, i):
         """Called with each ``draw_bits`` call, with ``i`` the index of the
-        corresonding block in ``self.examples.blocks``"""
+        corresponding block in ``self.examples.blocks``"""
 
     def stop_example(self, i, discarded):
         """Called at the end of each example, with ``i`` the
@@ -654,6 +655,9 @@ class _Overrun(object):
     def __repr__(self):
         return "Overrun"
 
+    def as_result(self):
+        return self
+
 
 Overrun = _Overrun()
 
@@ -711,6 +715,9 @@ class ConjectureResult(object):
         self.index = len(self.buffer)
         self.forced_indices = frozenset(self.forced_indices)
 
+    def as_result(self):
+        return self
+
 
 # Masks for masking off the first byte of an n-bit buffer.
 # The appropriate mask is stored at position n % 8.
@@ -721,24 +728,25 @@ BYTE_MASKS[0] = 255
 class ConjectureData(object):
     @classmethod
     def for_buffer(self, buffer, observer=None):
-        buffer = hbytes(buffer)
         return ConjectureData(
-            max_length=len(buffer),
-            draw_bytes=lambda data, n: hbytes(buffer[data.index : data.index + n]),
-            observer=observer,
+            prefix=buffer, max_length=len(buffer), parameter=None, observer=observer,
         )
 
-    def __init__(self, max_length, draw_bytes, observer=None):
+    def __init__(self, max_length, prefix, parameter, observer=None):
         if observer is None:
             observer = DataObserver()
         assert isinstance(observer, DataObserver)
+        self.__bytes_drawn = 0
         self.observer = observer
         self.max_length = max_length
         self.is_find = False
-        self._draw_bytes = draw_bytes
         self.overdraw = 0
         self.__block_starts = defaultdict(list)
         self.__block_starts_calculated_to = 0
+        self.__prefix = prefix
+        self.__parameter = parameter
+
+        assert parameter is not None or max_length <= len(prefix)
 
         self.blocks = Blocks(self)
         self.buffer = bytearray()
@@ -830,16 +838,22 @@ class ConjectureData(object):
                 % (strategy,)
             )
 
+        at_top_level = self.depth == 0
+        if at_top_level:
+            # We start this timer early, because accessing attributes on a LazyStrategy
+            # can be almost arbitrarily slow.  In cases like characters() and text()
+            # where we cache something expensive, this led to Flaky deadline errors!
+            # See https://github.com/HypothesisWorks/hypothesis/issues/2108
+            start_time = benchmark_time()
+
+        strategy.validate()
+
         if strategy.is_empty:
             self.mark_invalid()
 
         if self.depth >= MAX_DEPTH:
             self.mark_invalid()
 
-        return self.__draw(strategy, label=label)
-
-    def __draw(self, strategy, label):
-        at_top_level = self.depth == 0
         if label is None:
             label = strategy.label
         self.start_example(label=label)
@@ -849,7 +863,6 @@ class ConjectureData(object):
             else:
                 try:
                     strategy.validate()
-                    start_time = benchmark_time()
                     try:
                         return strategy.do_draw(self)
                     finally:
@@ -955,7 +968,6 @@ class ConjectureData(object):
 
         self.buffer = hbytes(self.buffer)
         self.events = frozenset(self.events)
-        del self._draw_bytes
         self.observer.conclude_test(self.status, self.interesting_origin)
 
     def draw_bits(self, n, forced=None):
@@ -971,9 +983,17 @@ class ConjectureData(object):
         self.__check_capacity(n_bytes)
 
         if forced is not None:
-            buf = bytearray(int_to_bytes(forced, n_bytes))
+            buf = int_to_bytes(forced, n_bytes)
+        elif self.__bytes_drawn < len(self.__prefix):
+            index = self.__bytes_drawn
+            buf = self.__prefix[index : index + n_bytes]
+            if len(buf) < n_bytes:
+                buf += uniform(self.__parameter.random, n_bytes - len(buf))
         else:
-            buf = bytearray(self._draw_bytes(self, n_bytes))
+            buf = self.__parameter.draw_bytes(n_bytes)
+        buf = bytearray(buf)
+        self.__bytes_drawn += n_bytes
+
         assert len(buf) == n_bytes
 
         # If we have a number of bits that is not a multiple of 8
@@ -1038,3 +1058,87 @@ def bits_to_bytes(n):
     Equivalent to (n + 7) // 8, but slightly faster. This really is
     called enough times that that matters."""
     return (n + 7) >> 3
+
+
+generation_parameters_count = 0
+
+
+class GenerationParameters(object):
+    """Parameters to control generation of examples."""
+
+    AVERAGE_ALPHABET_SIZE = 3
+
+    ALPHABET_FACTOR = math.log(1.0 - 1.0 / AVERAGE_ALPHABET_SIZE)
+
+    def __init__(self, random):
+        self.random = random
+        self.__pure_chance = None
+        self.__alphabet = {}
+
+        global generation_parameters_count
+        generation_parameters_count += 1
+
+        self.__id = generation_parameters_count
+
+    def __repr__(self):
+        return "GenerationParameters(%d)" % (self.__id,)
+
+    def draw_bytes(self, n):
+        """Draw an n-byte block from the distribution defined by this instance
+        of generation parameters."""
+        alphabet = self.alphabet(n)
+
+        if alphabet is None:
+            return self.__draw_without_alphabet(n)
+
+        return self.random.choice(alphabet)
+
+    def __draw_without_alphabet(self, n):
+        return uniform(self.random, n)
+
+    def alphabet(self, n_bytes):
+        """Returns an alphabet - a list of values to use for all blocks with
+        this number of bytes - or None if this value should be generated
+        without an alphabet.
+
+        This is designed to promote duplication in the test case that would
+        otherwise happen with very low probability.
+        """
+        try:
+            return self.__alphabet[n_bytes]
+        except KeyError:
+            pass
+
+        if self.random.random() <= self.pure_chance:
+            # Sometiems we don't want to use an alphabet (e.g. for generating
+            # sets of integers having a small alphabet is disastrous), so with
+            # some probability we want to generate choices that do not use the
+            # alphabet. As with other factors we set this probability globally
+            # across the whole choice of distribution so we have various levels
+            # of mixing.
+            result = None
+        else:
+            # We draw the size as a geometric distribution with average size
+            # GenerationParameters.AVERAGE_ALPHABET_SIZE.
+            size = (
+                int(
+                    math.log(self.random.random())
+                    / GenerationParameters.ALPHABET_FACTOR
+                )
+                + 1
+            )
+            assert size > 0
+
+            size = self.random.randint(1, 10)
+            result = [self.__draw_without_alphabet(n_bytes) for _ in hrange(size)]
+
+        self.__alphabet[n_bytes] = result
+        return result
+
+    @property
+    def pure_chance(self):
+        """Returns a probability with which any given draw_bytes call should
+        be forced to be all pure."""
+        if self.__pure_chance is None:
+            self.__pure_chance = self.random.random()
+        return self.__pure_chance
