@@ -17,7 +17,6 @@
 
 from __future__ import absolute_import, division, print_function
 
-import math
 from collections import Counter, defaultdict
 from enum import Enum
 from random import Random, getrandbits
@@ -28,16 +27,22 @@ import attr
 from hypothesis import HealthCheck, Phase, Verbosity, settings as Settings
 from hypothesis._settings import local_settings
 from hypothesis.internal.cache import LRUReusedCache
-from hypothesis.internal.compat import ceil, hbytes, hrange, int_from_bytes
+from hypothesis.internal.compat import ceil, hbytes, int_from_bytes
 from hypothesis.internal.conjecture.data import (
     ConjectureData,
     ConjectureResult,
+    GenerationParameters,
     Overrun,
     Status,
     StopTest,
 )
-from hypothesis.internal.conjecture.datatree import DataTree, TreeRecordingObserver
-from hypothesis.internal.conjecture.junkdrawer import uniform
+from hypothesis.internal.conjecture.datatree import (
+    DataTree,
+    PreviouslyUnseenBehaviour,
+    TreeRecordingObserver,
+)
+from hypothesis.internal.conjecture.junkdrawer import clamp
+from hypothesis.internal.conjecture.pareto import NO_SCORE, ParetoFront
 from hypothesis.internal.conjecture.shrinker import Shrinker, sort_key
 from hypothesis.internal.healthcheck import fail_health_check
 from hypothesis.reporting import base_report
@@ -47,9 +52,6 @@ from hypothesis.reporting import base_report
 __tracebackhide__ = True
 
 
-NO_SCORE = float("-inf")
-
-
 MAX_SHRINKS = 500
 CACHE_SIZE = 10000
 MUTATION_POOL_SIZE = 100
@@ -106,6 +108,17 @@ class ConjectureRunner(object):
         self.tree = DataTree()
 
         self.best_observed_targets = defaultdict(lambda: NO_SCORE)
+        self.best_examples_of_observed_targets = {}
+
+        # We keep the pareto front in the example database if we have one. This
+        # is only marginally useful at present, but speeds up local development
+        # because it means that large targets will be quickly surfaced in your
+        # testing.
+        if self.database_key is not None and self.settings.database is not None:
+            self.pareto_front = ParetoFront(self.random)
+            self.pareto_front.on_evict(self.on_pareto_evict)
+        else:
+            self.pareto_front = None
 
         # We want to be able to get the ConjectureData object that results
         # from running a buffer without recalculating, especially during
@@ -113,6 +126,10 @@ class ConjectureRunner(object):
         # executed test case.
         self.__data_cache = LRUReusedCache(CACHE_SIZE)
 
+    @property
+    def should_optimise(self):
+        return Phase.target in self.settings.phases
+
     def __tree_is_exhausted(self):
         return self.tree.is_exhausted
 
@@ -155,12 +172,30 @@ class ConjectureRunner(object):
 
         self.debug_data(data)
 
+        if self.pareto_front is not None and self.pareto_front.add(data.as_result()):
+            self.save_buffer(data.buffer, sub_key=b"pareto")
+
         assert len(data.buffer) <= BUFFER_SIZE
 
         if data.status >= Status.VALID:
             for k, v in data.target_observations.items():
                 self.best_observed_targets[k] = max(self.best_observed_targets[k], v)
 
+                if k not in self.best_examples_of_observed_targets:
+                    self.best_examples_of_observed_targets[k] = data.as_result()
+                    continue
+
+                existing_example = self.best_examples_of_observed_targets[k]
+                existing_score = existing_example.target_observations[k]
+
+                if v < existing_score:
+                    continue
+
+                if v > existing_score or sort_key(data.buffer) < sort_key(
+                    existing_example.buffer
+                ):
+                    self.best_examples_of_observed_targets[k] = data.as_result()
+
         if data.status == Status.VALID:
             self.valid_examples += 1
 
@@ -210,6 +245,9 @@ class ConjectureRunner(object):
 
         self.record_for_health_check(data)
 
+    def on_pareto_evict(self, data):
+        self.settings.database.delete(self.pareto_key, data.buffer)
+
     def generate_novel_prefix(self):
         """Uses the tree to proactively generate a starting sequence of bytes
         that we haven't explored yet for this test.
@@ -302,9 +340,9 @@ class ConjectureRunner(object):
                 HealthCheck.too_slow,
             )
 
-    def save_buffer(self, buffer):
+    def save_buffer(self, buffer, sub_key=None):
         if self.settings.database is not None:
-            key = self.database_key
+            key = self.sub_key(sub_key)
             if key is None:
                 return
             self.settings.database.save(key, hbytes(buffer))
@@ -313,13 +351,20 @@ class ConjectureRunner(object):
         if self.settings.database is not None and self.database_key is not None:
             self.settings.database.move(self.database_key, self.secondary_key, buffer)
 
-    @property
-    def secondary_key(self):
-        return b".".join((self.database_key, b"secondary"))
+    def sub_key(self, sub_key):
+        if self.database_key is None:
+            return None
+        if sub_key is None:
+            return self.database_key
+        return b".".join((self.database_key, sub_key))
 
     @property
-    def covering_key(self):
-        return b".".join((self.database_key, b"coverage"))
+    def secondary_key(self):
+        return self.sub_key(b"secondary")
+
+    @property
+    def pareto_key(self):
+        return self.sub_key(b"pareto")
 
     def note_details(self, data):
         self.__data_cache[data.buffer] = data.as_result()
@@ -425,29 +470,39 @@ class ConjectureRunner(object):
             )
             desired_size = max(2, ceil(0.1 * self.settings.max_examples))
 
-            for extra_key in [self.secondary_key, self.covering_key]:
-                if len(corpus) < desired_size:
-                    extra_corpus = list(self.settings.database.fetch(extra_key))
+            if len(corpus) < desired_size:
+                extra_corpus = list(self.settings.database.fetch(self.secondary_key))
 
-                    shortfall = desired_size - len(corpus)
+                shortfall = desired_size - len(corpus)
 
-                    if len(extra_corpus) <= shortfall:
-                        extra = extra_corpus
-                    else:
-                        extra = self.random.sample(extra_corpus, shortfall)
-                    extra.sort(key=sort_key)
-                    corpus.extend(extra)
+                if len(extra_corpus) <= shortfall:
+                    extra = extra_corpus
+                else:
+                    extra = self.random.sample(extra_corpus, shortfall)
+                extra.sort(key=sort_key)
+                corpus.extend(extra)
 
             for existing in corpus:
-                last_data = ConjectureData.for_buffer(
-                    existing, observer=self.tree.new_observer()
-                )
-                try:
-                    self.test_function(last_data)
-                finally:
-                    if last_data.status != Status.INTERESTING:
-                        self.settings.database.delete(self.database_key, existing)
-                        self.settings.database.delete(self.secondary_key, existing)
+                data = self.cached_test_function(existing)
+                if data.status != Status.INTERESTING:
+                    self.settings.database.delete(self.database_key, existing)
+                    self.settings.database.delete(self.secondary_key, existing)
+
+            # If we've not found any interesting examples so far we try some of
+            # the pareto front from the last run.
+            if len(corpus) < desired_size and not self.interesting_examples:
+                desired_extra = desired_size - len(corpus)
+                pareto_corpus = list(self.settings.database.fetch(self.pareto_key))
+                if len(pareto_corpus) > desired_extra:
+                    pareto_corpus = self.random.sample(pareto_corpus, desired_extra)
+                pareto_corpus.sort(key=sort_key)
+
+                for existing in pareto_corpus:
+                    data = self.cached_test_function(existing)
+                    if data not in self.pareto_front:
+                        self.settings.database.delete(self.pareto_key, existing)
+                    if data.status == Status.INTERESTING:
+                        break
 
     def exit_with(self, reason):
         self.debug("exit_with(%s)" % (reason.name,))
@@ -463,12 +518,12 @@ class ConjectureRunner(object):
             # the time to look for additional failures.
             return
 
+        self.debug("Generating new examples")
+
         zero_data = self.cached_test_function(hbytes(BUFFER_SIZE))
         if zero_data.status > Status.OVERRUN:
             self.__data_cache.pin(zero_data.buffer)
 
-        self.optimise_all(zero_data)
-
         if zero_data.status == Status.OVERRUN or (
             zero_data.status == Status.VALID and len(zero_data.buffer) * 2 > BUFFER_SIZE
         ):
@@ -490,14 +545,6 @@ class ConjectureRunner(object):
         self.health_check_state = HealthCheckState()
 
         def should_generate_more():
-            # If we haven't found a bug, keep looking.  We check this before
-            # doing anything else as it's by far the most common case.
-            if not self.interesting_examples:
-                return True
-            # If we've found a bug and won't report more than one, stop looking.
-            elif not self.settings.report_multiple_bugs:
-                return False
-            assert self.first_bug_found_at <= self.last_bug_found_at <= self.call_count
             # End the generation phase where we would have ended it if no bugs had
             # been found.  This reproduces the exit logic in `self.test_function`,
             # but with the important distinction that this clause will move on to
@@ -506,8 +553,23 @@ class ConjectureRunner(object):
             if (
                 self.valid_examples >= self.settings.max_examples
                 or self.call_count >= max(self.settings.max_examples * 10, 1000)
+                or (
+                    self.best_examples_of_observed_targets
+                    and self.valid_examples * 2 >= self.settings.max_examples
+                    and self.should_optimise
+                )
             ):  # pragma: no cover
                 return False
+
+            # If we haven't found a bug, keep looking - if we hit any limits on
+            # the number of tests to run that will raise an exception and stop
+            # the run.
+            if not self.interesting_examples:
+                return True
+            # If we've found a bug and won't report more than one, stop looking.
+            elif not self.settings.report_multiple_bugs:
+                return False
+            assert self.first_bug_found_at <= self.last_bug_found_at <= self.call_count
             # Otherwise, keep searching for between ten and 'a heuristic' calls.
             # We cap 'calls after first bug' so errors are reported reasonably
             # soon even for tests that are allowed to run for a very long time,
@@ -534,14 +596,103 @@ class ConjectureRunner(object):
         parameter = GenerationParameters(self.random)
         count = 0
 
+        # We attempt to use the size of the minimal generated test case starting
+        # from a given novel prefix as a guideline to generate smaller test
+        # cases for an initial period, by restriscting ourselves to test cases
+        # that are not much larger than it.
+        #
+        # Calculating the actual minimal generated test case is hard, so we
+        # take a best guess that zero extending a prefix produces the minimal
+        # test case starting with that prefix (this is true for our built in
+        # strategies). This is only a reasonable thing to do if the resulting
+        # test case is valid. If we regularly run into situations where it is
+        # not valid then this strategy is a waste of time, so we want to
+        # abandon it early. In order to do this we track how many times in a
+        # row it has failed to work, and abort small test case generation when
+        # it has failed too many times in a row.
+        consecutive_zero_extend_is_invalid = 0
+
         while should_generate_more():
             prefix = self.generate_novel_prefix()
+            assert len(prefix) <= BUFFER_SIZE
+
+            # We control growth during initial example generation, for two
+            # reasons:
+            #
+            # * It gives us an opportunity to find small examples early, which
+            #   gives us a fast path for easy to find bugs.
+            # * It avoids low probability events where we might end up
+            #   generating very large examples during health checks, which
+            #   on slower machines can trigger HealthCheck.too_slow.
+            #
+            # The heuristic we use is that we attempt to estimate the smallest
+            # extension of this prefix, and limit the size to no more than
+            # an order of magnitude larger than that. If we fail to estimate
+            # the size accurately, we skip over this prefix and try again.
+            #
+            # We need to tune the example size based on the initial prefix,
+            # because any fixed size might be too small, and any size based
+            # on the strategy in general can fall afoul of strategies that
+            # have very different sizes for different prefixes.
+            small_example_cap = clamp(10, self.settings.max_examples // 10, 50)
+
+            if (
+                self.valid_examples <= small_example_cap
+                and self.call_count <= 5 * small_example_cap
+                and not self.interesting_examples
+                and consecutive_zero_extend_is_invalid < 5
+            ):
+                minimal_example = self.cached_test_function(
+                    prefix + hbytes(BUFFER_SIZE - len(prefix))
+                )
+
+                if minimal_example.status < Status.VALID:
+                    consecutive_zero_extend_is_invalid += 1
+                    continue
+
+                consecutive_zero_extend_is_invalid = 0
+
+                minimal_extension = len(minimal_example.buffer) - len(prefix)
+
+                max_length = min(len(prefix) + minimal_extension * 10, BUFFER_SIZE)
+
+                # We could end up in a situation where even though the prefix was
+                # novel when we generated it, because we've now tried zero extending
+                # it not all possible continuations of it will be novel. In order to
+                # avoid making redundant test calls, we rerun it in simulation mode
+                # first. If this has a predictable result, then we don't bother
+                # running the test function for real here. If however we encounter
+                # some novel behaviour, we try again with the real test function,
+                # starting from the new novel prefix that has discovered.
+                try:
+                    trial_data = self.new_conjecture_data(
+                        prefix=prefix, parameter=parameter, max_length=max_length
+                    )
+                    self.tree.simulate_test_function(trial_data)
+                    continue
+                except PreviouslyUnseenBehaviour:
+                    pass
+
+                # If the simulation entered part of the tree that has been killed,
+                # we don't want to run this.
+                if trial_data.observer.killed:
+                    continue
+
+                # We might have hit the cap on number of examples we should
+                # run when calculating the minimal example.
+                if not should_generate_more():
+                    break
+
+                prefix = trial_data.buffer
+            else:
+                max_length = BUFFER_SIZE
+
+            data = self.new_conjecture_data(
+                prefix=prefix, parameter=parameter, max_length=max_length
+            )
 
-            data = self.new_conjecture_data(draw_bytes_with(prefix, parameter))
             self.test_function(data)
 
-            self.optimise_all(data)
-
             count += 1
             if (
                 data.status < Status.VALID
@@ -551,23 +702,32 @@ class ConjectureRunner(object):
                 count = 0
                 parameter = GenerationParameters(self.random)
 
-    def optimise_all(self, data):
-        """If the data or result object is suitable for hill climbing, run hill
-        climbing on all of its target observations."""
-        if data.status == Status.VALID:
-            for target in data.target_observations:
+    def optimise_targets(self):
+        """If any target observations have been made, attempt to optimise them
+        all."""
+        if not self.should_optimise:
+            return
+        while True:
+            prev_calls = self.call_count
+            for target, data in list(self.best_examples_of_observed_targets.items()):
                 self.new_optimiser(data, target).run()
+            # This shows up as a missing branch for some reason but is definitely
+            # covered if you add an assertion to check.
+            if self.interesting_examples or (prev_calls == self.call_count):
+                break
 
     def _run(self):
         self.reuse_existing_examples()
         self.generate_new_examples()
+        self.optimise_targets()
         self.shrink_interesting_examples()
         self.exit_with(ExitReason.finished)
 
-    def new_conjecture_data(self, draw_bytes):
+    def new_conjecture_data(self, prefix, parameter, max_length=BUFFER_SIZE):
         return ConjectureData(
-            draw_bytes=draw_bytes,
-            max_length=BUFFER_SIZE,
+            prefix=prefix,
+            parameter=parameter,
+            max_length=max_length,
             observer=self.tree.new_observer(),
         )
 
@@ -723,135 +883,3 @@ class ConjectureRunner(object):
         result = str(event)
         self.events_to_strings[event] = result
         return result
-
-
-generation_parameters_count = 0
-
-
-class GenerationParameters(object):
-    """Parameters to control generation of examples."""
-
-    AVERAGE_ALPHABET_SIZE = 3
-
-    ALPHABET_FACTOR = math.log(1.0 - 1.0 / AVERAGE_ALPHABET_SIZE)
-
-    def __init__(self, random):
-        self.__random = random
-        self.__zero_chance = None
-        self.__max_chance = None
-        self.__pure_chance = None
-        self.__alphabet = {}
-
-        global generation_parameters_count
-        generation_parameters_count += 1
-
-        self.__id = generation_parameters_count
-
-    def __repr__(self):
-        return "GenerationParameters(%d)" % (self.__id,)
-
-    def draw_bytes(self, n):
-        """Draw an n-byte block from the distribution defined by this instance
-        of generation parameters."""
-        alphabet = self.alphabet(n)
-
-        if alphabet is None:
-            return self.__draw_without_alphabet(n)
-
-        return self.__random.choice(alphabet)
-
-    def __draw_without_alphabet(self, n):
-        if self.__random.random() <= self.zero_chance:
-            return hbytes(n)
-
-        if self.__random.random() <= self.max_chance:
-            return hbytes([255]) * n
-
-        return uniform(self.__random, n)
-
-    def alphabet(self, n_bytes):
-        """Returns an alphabet - a list of values to use for all blocks with
-        this number of bytes - or None if this value should be generated
-        without an alphabet.
-
-        This is designed to promote duplication in the test case that would
-        otherwise happen with very low probability.
-        """
-        try:
-            return self.__alphabet[n_bytes]
-        except KeyError:
-            pass
-
-        if self.__random.random() <= self.pure_chance:
-            # Sometiems we don't want to use an alphabet (e.g. for generating
-            # sets of integers having a small alphabet is disastrous), so with
-            # some probability we want to generate choices that do not use the
-            # alphabet. As with other factors we set this probability globally
-            # across the whole choice of distribution so we have various levels
-            # of mixing.
-            result = None
-        else:
-            # We draw the size as a geometric distribution with average size
-            # GenerationParameters.AVERAGE_ALPHABET_SIZE.
-            size = (
-                int(
-                    math.log(self.__random.random())
-                    / GenerationParameters.ALPHABET_FACTOR
-                )
-                + 1
-            )
-            assert size > 0
-
-            size = self.__random.randint(1, 10)
-            result = [self.__draw_without_alphabet(n_bytes) for _ in hrange(size)]
-
-        self.__alphabet[n_bytes] = result
-        return result
-
-    @property
-    def max_chance(self):
-        """Returns a probability with which any given draw_bytes call should
-        be forced to be all 255 bytes. This is an important value because it
-        can make it more likely to push us into rare corners of the search
-        space, especially when biased_coin is being used."""
-        if self.__max_chance is None:
-            # We want to generate pure random examples every now and then. This
-            # is partly to offset too strong a bias to zero and partly because
-            # some user defined strategies may not play well with zero biasing.
-            self.__max_chance = self.__random.random() * 0.01
-        return self.__max_chance
-
-    @property
-    def zero_chance(self):
-        """Returns a probability with which any given draw_bytes call should
-        be forced to be all zero. This is an important value especially because
-        it will tend to force the test case to be smaller than it otherwise
-        would be."""
-        if self.__zero_chance is None:
-            # We want to generate pure random examples every now and then. This
-            # is partly to offset too strong a bias to zero and partly because
-            # some user defined strategies may not play well with zero biasing.
-            if self.__random.randrange(0, 10) == 0:
-                self.__zero_chance = 0.0
-            else:
-                self.__zero_chance = self.__random.random() * 0.5
-        return self.__zero_chance
-
-    @property
-    def pure_chance(self):
-        """Returns a probability with which any given draw_bytes call should
-        be forced to be all pure."""
-        if self.__pure_chance is None:
-            self.__pure_chance = self.__random.random()
-        return self.__pure_chance
-
-
-def draw_bytes_with(prefix, parameter):
-    def draw_bytes(data, n):
-        if data.index < len(prefix):
-            result = prefix[data.index : data.index + n]
-            # We always draw prefixes as a whole number of blocks
-            return result
-        return parameter.draw_bytes(n)
-
-    return draw_bytes
